/*
************************************************************************************
*                                                                                  * 
*  Copyright (c) 2004, Filipe Maia                                                 *
*   All rights reserved.                                                           *
*                                                                                  *
*  Redistribution and use in source and binary forms, with or without              *
*  modification, are permitted provided that the following conditions are met:     *
*    -Redistributions of source code must retain the above copyright notice, this  *
*     list of conditions and the following disclaimer.                             *
*    -Redistributions in binary form must reproduce the above copyright notice,    *
*     this list of conditions and the following disclaimer in the documentation    *
*     and/or other materials provided with the distribution.                       *
*    -The names of its contributors may not be used to endorse or promote          *
*     products derived from this software without specific prior written           *
*     permission.                                                                  *
*                                                                                  *
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"     *
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE       *
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE  *
*  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE     *
*  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL      *
*  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR      *
*  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER      *
*  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,   *
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   *
*  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.            *
*                                                                                  *
************************************************************************************
*/

#ifndef _BITBOARD_H_
#define _BITBOARD_H_

#define _GNU_SOURCE
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <inttypes.h>



#define mask(A) mask[A]

#define ROW(x) (x >> 3)
#define RANK(x) (7 - (x >> 3))
#define COL(x) (x & 7)
/* As i've become sane again I'll use H1 for *MOST* significant bit 
   and A8 for least significant bit so the board looks something like this

   BBBBBBBB 0000000
   BBBBBBBB 0000000
   ........ 0000000
   ........ 0000000  <- A really big number
   ........ 0000000
   ........ 0000000
   WWWWWWWW 0000000
   WWWWWWWW 0000001 

*/

typedef uint64_t BitBoard;

#define BPawns game.pieces[5]
#define BRooks game.pieces[4]
#define BKnights game.pieces[3]
#define BBishops game.pieces[2]
#define BQueens game.pieces[1]
#define BKing game.pieces[0]
#define WPawns game.pieces[7]
#define WRooks game.pieces[8]
#define WKnights game.pieces[9]
#define WBishops game.pieces[10]
#define WQueens game.pieces[11]
#define WKing game.pieces[12]


typedef enum {none=0, 
	      w_pawn=1, w_rook,w_knight,w_bishop,w_queen,w_king=6,
	      b_king=-6,b_queen,b_bishop,b_knight,b_rook, b_pawn=-1}Colour_Piece;

typedef enum {NONE=0, 
	      PAWN=1, ROOK,KNIGHT,BISHOP,QUEEN,KING=6}Piece;


typedef struct{
  BitBoard zob_key;
  BitBoard pawn_key;
  char to_move;
  /* This is set to en_passant target 
   0 means that there is no enpassant target */
  char en_passant;
  unsigned char new_castle;
  char comp;
  /* positive for white */
  Colour_Piece board[64];

  int black_material;
  int white_material;
  int black_majors;
  int white_majors;
  int white_minors;
  int black_minors;
  int white_pawns;
  int black_pawns;

  int white_king;
  int black_king;

  /* Now comes the heavy stuff */

/*
 **********************************************
 *                                            *
 *  this array of bitboards correspondes to   * 
 *  0 = b_king, 1 = b_queen, 2 = b_bishop...  *
 *  , 11 = w_queen, 12 = w_king               *
 *                                            *
 **********************************************
 */
  BitBoard pieces[13];

  BitBoard b_pieces;
  BitBoard w_pieces;

  BitBoard all_r_90;
  BitBoard all_even_diag;
  BitBoard all_odd_diag;

  BitBoard passed_pawns;
} Position;




extern const int even_diag_shift[64];
extern const int odd_diag_shift[64];
extern const int len_even_diag[64];
extern const int len_odd_diag[64];
extern const int r_90_map[64];
extern const int r_90_shift[64];
extern const int r_0_shift[64];
extern const int even_diag_map[64];
extern const int odd_diag_map[64];

extern BitBoard knight_moves[64];
extern BitBoard king_moves[64];
extern BitBoard rook_rank_moves[64][256];
extern BitBoard rook_file_moves[64][256];
extern BitBoard bishop_odd_moves[64][256];
extern BitBoard bishop_even_moves[64][256];




extern const signed char white_outpost[64];


extern int passed_scale;
extern const char outside_passed[128];

extern  signed char kval_wn[64];
extern  signed char kval_wk[64];
extern  signed char kval_wq[64];

void init_w_pawns(BitBoard * p);
void init_b_pawns(BitBoard * p);
void init_b_knights(BitBoard * p);
void init_w_knights(BitBoard * p);
void init_b_bishops(BitBoard * p);
void init_w_bishops(BitBoard * p);
void init_b_rooks(BitBoard * p);
void init_w_rooks(BitBoard * p);
void init_b_queens(BitBoard *p);
void init_w_queens(BitBoard *p);
void init_queen_bishops(BitBoard * p);
void init_queen_rooks(BitBoard * p);
void init_all_r_90(BitBoard * p);
void init_all_even_diag(BitBoard * p);
void init_all_odd_diag(BitBoard * p);


#ifdef _DEBUG
void make_const_boards_clones();
int check_board_corruption();
#endif

#endif /* ifndef _BITBOARD_H_*/
