/*
************************************************************************************
*                                                                                  * 
*  Copyright (c) 2004, Filipe Maia                                                 *
*   All rights reserved.                                                           *
*                                                                                  *
*  Redistribution and use in source and binary forms, with or without              *
*  modification, are permitted provided that the following conditions are met:     *
*    -Redistributions of source code must retain the above copyright notice, this  *
*     list of conditions and the following disclaimer.                             *
*    -Redistributions in binary form must reproduce the above copyright notice,    *
*     this list of conditions and the following disclaimer in the documentation    *
*     and/or other materials provided with the distribution.                       *
*    -The names of its contributors may not be used to endorse or promote          *
*     products derived from this software without specific prior written           *
*     permission.                                                                  *
*                                                                                  *
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"     *
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE       *
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE  *
*  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE     *
*  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL      *
*  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR      *
*  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER      *
*  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,   *
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   *
*  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.            *
*                                                                                  *
************************************************************************************
*/
#include <stdio.h>

#include <bitboard.h>
#include <init.h>
#include <gen_moves.h>
#include <chess_const.h>
#include <validate.h>
#include <math_util.h>
#include <bitboard.h>
#include <score.h>
#include <init_moves.h>
#include <zobrist.h>
#include "search.h"
#include <tt.h>
#include <controls.h>

/* And here is the almighty Position */
Position game;

/*
  If it's possible for a queen to move from a to b
  it returns a bitboard with the squares it has to travel.
  Otherwise returns 0 (!notice that 0 is also returned when
  a and b are next to eachother!)
*/
BitBoard path[64][64];

/* 
   Returns the direction of the vector between 2 squares
   if they are on the same line/diagonal. Otherwise returns 0
 */
short directions[64][64];
/* beam is a bitboard that contains a ray directed 
 from a to b and that starts right after b 
              a    b                       */

BitBoard beam[64][64];


/* returns the first n bits turned on*/
BitBoard n_bit_mask[65];

/* Returns the a Bitboard with the bit n turned on */
BitBoard mask[64];
/* Returns the a Bitboard with the rank n+1 (meaning mask_rank[0] will mask rank 1 in chess nomenclature) turned on */
BitBoard mask_rank[8];
/*
  Returns the a Bitboard with the file n turned on 
  0=A,1=B...7=H
*/
BitBoard mask_file[8];

/* 
   This is pretty much like the pawn attacks boards
   expect that it's only for the en passant squares and
   both colors on the same board
*/
BitBoard en_passant_mask[64];

BitBoard mask_qr_trapped_b[3];
BitBoard mask_kr_trapped_b[3];
BitBoard mask_qr_trapped_w[3];
BitBoard mask_kr_trapped_w[3];


/*
  Contains the squares attacked be a pawn
  on each of the 64 squares 
*/
BitBoard b_pawn_attacks[64];
BitBoard w_pawn_attacks[64];

BitBoard light_squares;
BitBoard dark_squares;

BitBoard mask_WBT;
BitBoard mask_BBT;

BitBoard mask_abc;
BitBoard mask_fgh;
BitBoard mask_abcd;
BitBoard mask_efgh;

signed char kval_bn[64];
signed char kval_bk[64];
signed char kval_bq[64];


void init_const_boards(){
  /* init random generator */

  /*  srandom(time(NULL));*/
  init_masks();
  init_knight_moves();
  init_king_moves();
  init_rook_rank_moves();
  init_rook_file_moves();
  init_bishop_even_moves();
  init_bishop_odd_moves();
  init_zobrist();
  init_val_table();
  init_path();
  init_beam();
#ifdef _DEBUG  
  make_const_boards_clones();
#endif
}


void init_masks(){
  int i,j;
  int sq;
  int bishopsq[4] = {+7,+9,-7,-9};
  BitBoard a = 1;
  for(i = 0;i<64;i++){
    mask[i] = a;
    a = a << 1;
  } 
  a = 0xff;
  for(i = 0;i<8;i++){
    mask_rank[i] = a << ((7-i)*8);
  }
  a = 0x8080808080808080ULL;
  for(i = 0;i<8;i++){
    mask_file[i] = a >> (i);
  }
  n_bit_mask[0] = 0;
  for(i = 1;i<65;i++){
    n_bit_mask[i] = (n_bit_mask[i-1] << 1) + 1;
  }
  
  mask_qr_trapped_b[0]=mask[A7];
  mask_qr_trapped_b[1]=mask[A8] | mask[A7];
  mask_qr_trapped_b[2]=mask[A8] | mask[B8] | mask[A7];

  mask_kr_trapped_b[0]=mask[H7];
  mask_kr_trapped_b[1]=mask[H8] | mask[H7];
  mask_kr_trapped_b[2]=mask[H8] | mask[G8] | mask[H7];
  
  mask_qr_trapped_w[0]=mask[A2];
  mask_qr_trapped_w[1]=mask[A1] | mask[A2];
  mask_qr_trapped_w[2]=mask[A1] | mask[B1] | mask[A2];
  
  mask_kr_trapped_w[0]=mask[H2];
  mask_kr_trapped_w[1]=mask[H1] | mask[H2];
  mask_kr_trapped_w[2]=mask[H1] | mask[G1] | mask[H2];
  
  a = 1;
  
  for(i=0;i<64;i++) {
    w_pawn_attacks[i]=0;
    if (i < 56)
      for(j=2;j<4;j++) {
        sq=i+bishopsq[j];
        if((abs(RANK(sq)-RANK(i))==1) && 
           (abs(COL(sq) - COL(i))==1) &&
	   (sq < 64) && (sq > -1)) 
          w_pawn_attacks[i]=w_pawn_attacks[i] | a<<sq;
      }
    b_pawn_attacks[i]=0;
    if (i > 7)
      for(j=0;j<2;j++) {
        sq=i+bishopsq[j];
        if((abs(RANK(sq)-RANK(i))==1) && 
           (abs(COL(sq)-COL(i))==1) &&
	   (sq < 64) && (sq > -1)) 
          b_pawn_attacks[i]=b_pawn_attacks[i] | a<<sq;
      }
  }
  
  for (i=0;i<64;i++) {
    if (i&1) {
      light_squares=light_squares | a<<i;
    }
    else {
      dark_squares=dark_squares | a<<i;
    }
  }

  mask_abc = mask_file[FILEA] | mask_file[FILEB] | mask_file[FILEC];
  mask_abcd = mask_file[FILEA] | mask_file[FILEB] | mask_file[FILEC] | mask_file[FILED];
  mask_fgh = mask_file[FILEF] | mask_file[FILEG] | mask_file[FILEH];
  mask_efgh = mask_file[FILEE] | mask_file[FILEF] | mask_file[FILEG] | mask_file[FILEH];

  mask_WBT = mask(A7) | mask(B8) | mask(H7) | mask(G8);
  mask_BBT = mask(A2) | mask(B1) | mask(H2) | mask(G1);

  for (i = 0; i < 8; i++){
    for (j = 0; j < 8; j++) {
      kval_bn[i * 8 + j] = kval_wn[(7 - i) * 8 + j];
      kval_bk[i * 8 + j] = kval_wk[(7 - i) * 8 + j];
      kval_bq[i * 8 + j] = kval_wq[(7 - i) * 8 + j];
    }
  }

  for (i = 0; i < 64; i++)
    en_passant_mask[i] = 0;
  for (i = B6; i <= G6; i++)
    en_passant_mask[i] = mask[i-1+8] | mask[i+1+8];
  for (i = B3; i <= G3; i++)
    en_passant_mask[i] = mask[i-1-8] | mask[i+1-8];
  en_passant_mask[A3] = mask[B4];
  en_passant_mask[H3] = mask[G4];
  en_passant_mask[A6] = mask[B5];
  en_passant_mask[H6] = mask[G5];
}


void init_game(){
  int i;
  BitBoard a;
  init_tt();
  init_w_pawns(&WPawns);
  init_b_pawns(&BPawns);
  init_w_bishops(&WBishops);
  init_b_bishops(&BBishops);
  init_w_knights(&WKnights);
  init_b_knights(&BKnights);
  init_w_rooks(&WRooks);
  init_b_rooks(&BRooks);
  init_w_queens(&WQueens);
  init_b_queens(&BQueens);
  init_all_r_90(&(game.all_r_90));
  init_all_even_diag(&(game.all_even_diag));
  init_all_odd_diag(&(game.all_odd_diag));


  game.comp = BLACK;
  game.to_move = WHITE;
  game.en_passant = 0;
  game.new_castle = W_BIG_CASTLE|W_SMALL_CASTLE | B_BIG_CASTLE |B_SMALL_CASTLE;

  game.white_material = QUEEN_VALUE+ROOK_VALUE*2+BISHOP_VALUE*2+KNIGHT_VALUE*2+PAWN_VALUE*8;
  game.black_material = QUEEN_VALUE+ROOK_VALUE*2+BISHOP_VALUE*2+KNIGHT_VALUE*2+PAWN_VALUE*8;
  game.white_majors = 3;
  game.black_majors = 3;
  game.white_minors = 4;
  game.black_minors = 4;
  game.white_pawns = 8;
  game.black_pawns = 8;
  game.white_king = 60;
  game.black_king = 4;

  
  for(i = 0;i<64;i++){
    game.board[i] = none;
  }
  for(i = 0;i<8;i++){
    game.board[i+8] = b_pawn;
    game.board[i+48] = w_pawn;
  }
  game.board[A8] = game.board[H8] = b_rook;
  game.board[A1] = game.board[H1] = w_rook;
  game.board[B8] = game.board[G8] = b_knight;
  game.board[G1] = game.board[B1] = w_knight;
  game.board[C8] = game.board[F8] = b_bishop;
  game.board[F1] = game.board[C1] = w_bishop;
  game.board[D8] = b_queen;
  game.board[D1] = w_queen;
  game.board[E8] = b_king;
  game.board[E1] = w_king;

  a = 1;
  game.w_pieces = WPawns | WKnights | WRooks |
    WBishops | WQueens | a << (game.white_king);
  a = 1;
  game.b_pieces = BPawns | BKnights | BRooks |
    BBishops | BQueens | a << (game.black_king);
  
  game.pawn_key = gen_pawn_key();
  game.zob_key = gen_zob_key();
  g_control.moves_out_of_book = 0;
  g_control.move_nr = 0;
  g_control.opponent[0] = 0;
  g_control.elo[0] = 0;
  g_control.elo[1] = 0;
  if(g_control.pgn)
    fclose(g_control.pgn);
  g_control.pgn = fopen("games.pgn","a");
  setvbuf(g_control.pgn,(char *)NULL, _IOLBF, 0);
}


void init_path(){
  int i,j,k;
  BitBoard a;
  init_directions();
  for(i = 0;i<64;i++){
    for(j = 0;j<64;j++){     
      if(abs(distance(i,j)) == 1){
	a = 0;
	path[i][j] = a;
      }else{
	if(directions[i][j] == RANK_RIGHT){
	  k = i +1;
	  a = mask[k];
	  while(++k != j){
	    a |= mask[k];
	  }
	}else if(directions[i][j] == RANK_LEFT){
	  k = i - 1;
	  a = mask[k];
	  while(--k != j){
	    a |= mask[k];
	  }
	}else if(directions[i][j] == FILE_UP){
	  k = i - 8;
	  a = mask[k];
	  while(k-8 != j){
	    k -= 8;
	    a |= mask[k];
	  }
	}else if(directions[i][j] == FILE_DOWN){
	  k = i + 8;
	  a = mask[k];
	  while(k+8 != j){
	    k += 8;
	    a |= mask[k];
	  }
	}else if(directions[i][j] == ODD_UP){
	  k = i - 7;
	  a = mask[k];
	  while(k-7 != j){
	    k -= 7;
	    a |= mask[k];
	  }
	}else if(directions[i][j] == ODD_DOWN){
	  k = i + 7;
	  a = mask[k];
	  while(k+7 != j){
	    k += 7;
	    a |= mask[k];
	  }
	}else if(directions[i][j] == EVEN_UP){
	  k = i - 9;
	  a = mask[k];
	  while(k-9 != j){
	    k -= 9;
	    a |= mask[k];
	  }
	}else if(directions[i][j] == EVEN_DOWN){
	  k = i + 9;
	  a = mask[k];
	  while(k+9 != j){
	    k += 9;
	    a |= mask[k];
	  }
	}else{
	  a = 0;
	}
	path[i][j] = a;
      }
    }
  }
}

void init_directions(){
  int from,to;
  for(from = 0;from<64;from++){
    for(to = 0;to<64;to++){     
      if(ROW(from) == ROW(to)){
	if(from < to){
	  directions[from][to] = RANK_RIGHT;
	}else if(to < from){
	  directions[from][to] = RANK_LEFT;
	}else{
	  directions[from][to] = 0;
	}
      }else if(COL(from) == COL(to)){
	if(from > to){
	  directions[from][to] = FILE_UP;
	}else if(to > from){
	  directions[from][to] = FILE_DOWN;
	}else{
	  directions[from][to] = 0;
	}
      }else if(ROW(from)-ROW(to) == COL(from) - COL(to)){
	if(from > to){
	  directions[from][to] = EVEN_UP;
	}else if(to > from){
	  directions[from][to] = EVEN_DOWN;
	}else{
	  directions[from][to] = 0;
	}
      }else if(ROW(from)-ROW(to) == -(COL(from) - COL(to))){
	if(from > to){
	  directions[from][to] = ODD_UP;
	}else if(to > from){
	  directions[from][to] = ODD_DOWN;
	}else{
	  directions[from][to] = 0;
	}
      }else{
	directions[from][to] = 0;
      }
    }
  }
}

void init_beam(){
  int a,b;
  BitBoard board ;
  int i;
  for(a = 0;a<64;a++){
    for(b = 0;b<64;b++){
      board = 0;
	if(directions[a][b] == RANK_RIGHT){
	  for(i = b+1;COL(i) != 0;i++)
	    board |= mask[i];
	}else if(directions[a][b] == RANK_LEFT){
	  for(i = b-1;COL(i) != 7;i--)
	    board |= mask[i];
	}else if(directions[a][b] == FILE_UP){
	  for(i = b-8;i >= 0;i -= 8)
	    board |= mask[i];
	}else if(directions[a][b] == FILE_DOWN){
	  for(i = b+8;i < 64;i += 8)
	    board |= mask[i];
	}else if(directions[a][b] == ODD_UP){
	  for(i = b-7;i >= 0 && COL(i) != 0;i -= 7)
	    board |= mask[i];
	}else if(directions[a][b] == ODD_DOWN){
	  for(i = b+7;i < 64 && COL(i) != 7;i += 7)
	    board |= mask[i];
	}else if(directions[a][b] == EVEN_UP){
	  for(i = b-9;i >= 0 && COL(i) != 7;i -= 9)
	    board |= mask[i];
	}else if(directions[a][b] == EVEN_DOWN){
	  for(i = b+9;i < 64 && COL(i) != 0;i += 9)
	    board |= mask[i];
	}
	beam[a][b] = board;
    }
  }
}
