/*
************************************************************************************
*                                                                                  * 
*  Copyright (c) 2004, Filipe Maia                                                 *
*   All rights reserved.                                                           *
*                                                                                  *
*  Redistribution and use in source and binary forms, with or without              *
*  modification, are permitted provided that the following conditions are met:     *
*    -Redistributions of source code must retain the above copyright notice, this  *
*     list of conditions and the following disclaimer.                             *
*    -Redistributions in binary form must reproduce the above copyright notice,    *
*     this list of conditions and the following disclaimer in the documentation    *
*     and/or other materials provided with the distribution.                       *
*    -The names of its contributors may not be used to endorse or promote          *
*     products derived from this software without specific prior written           *
*     permission.                                                                  *
*                                                                                  *
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"     *
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE       *
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE  *
*  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE     *
*  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL      *
*  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR      *
*  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER      *
*  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,   *
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   *
*  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.            *
*                                                                                  *
************************************************************************************
*/

#include <gen_moves.h>
#include <chess_const.h>
#include "bitboard.h"
#include "search.h"
#include "util.h"
#include "validate.h"
#include "consistency.h"
#include "score.h"
#include "make_moves.h"
#include <init.h>
#include <zobrist.h>


/* our move counter */
unsigned char move_50_counter;

/* This is our history stack */
BitBoard move_history[MAX_HISTORY*2];

/* pointer to the move right after the last irreversible change */
BitBoard * history_floor;

/* turn counter */
int turn;

static Move_Data data_stack [MAX_DEPTH];

Move_Data * data_stack_top;


void init_move_maker(Position * p){
  move_50_counter = 0;
  move_history[0] = p->zob_key;
  history_floor = &(move_history[0]);
  reset_data_stack();
}
  
void reset_data_stack(){
  data_stack_top = &(data_stack[0]);
}


void make_move(Move m){
  Move_Data * md = data_stack_top;
  int sq1 = 0;
  int sq2 = 0;
  int sq3 = 0;
  int sq4 = 0;
#ifdef _DEBUG
  FILE * fd;
  Move debug_stack[MAX_POSSIBLE_MOVES];
  Move * debug_top = &debug_stack[0];
  Move * debug_floor = debug_top;
  char debug_flag = 0;
#endif

#ifdef _DEBUG

  debug_top = debug_floor;
  debug_flag = 1;
  if(game.to_move == WHITE){
    debug_top = all_w_non_cap(debug_top);
  }else{
    debug_top = all_b_non_cap(debug_top);
  }
  if(game.to_move == WHITE){
    debug_top = all_w_cap(debug_top);
  }else{
    debug_top = all_b_cap(debug_top);
  }
  while(debug_top>debug_floor){
    debug_top--;
    if(*((int *)(debug_top)) == *((int*)&m)){
	  debug_flag = 0;
    }
  }  
  if(debug_flag){
    fd = fopen("error.log","a");
    print_board(fd);
    print_board(stdout);
    fclose(fd);
    print_moves(&m+1,&m);
    error("Move not really valid!");
  }
  
  if( m.from > 63){
    error("Big Error!");
  }
  if( m.to > 63){
    error("Big Error!");
  } 
    

  if(!validate(m) || !check_boards_cons()){
    print_board(stdout);
    print_moves(&m+1,&m);
    error("Insane move!");
  }
  if(gen_zob_key() != game.zob_key || gen_pawn_key() != game.pawn_key){
    print_board(stdout);
    print_moves(&m+1,&m);
    error("Zobrist will be really mad!");
  }
  

#endif

  /* save data code */
  md->new_castle = game.new_castle;
  md->move_50_counter = move_50_counter;
  md->history_floor  = history_floor;
  md->zob_key = game.zob_key;
  md->pawn_key = game.pawn_key;
  md->en_passant = game.en_passant;

  /* update move_counter that also serves as turn counter*/
  move_50_counter++;

  /* "switch off" previous en_passant */
  if(game.en_passant){
    game.zob_key ^= en_passant_zob[(int)game.en_passant];
    game.en_passant = 0;
  }

  /* we don't have to switch off castling because you can never regain
     castling availability with a move 
  */

  /* capture code */
  if(m.special & CAPTURE){
    /* en passant code*/
    if(m.special & EN_PASSANT){
      if(game.board[m.from] < 0){
	/*target is white */
	game.white_material -= PAWN_VALUE;
	WPawns ^= mask[m.to-8];
	game.w_pieces ^= mask[m.to-8];
	game.zob_key ^= zobrist[PAWN-1][WHITE][m.to-8];
	game.pawn_key ^= zobrist[PAWN-1][WHITE][m.to-8];
	game.all_r_90 ^= mask[r_90_map[m.to-8]];
	game.all_odd_diag ^= mask[odd_diag_map[m.to-8]];
	game.all_even_diag ^= mask[even_diag_map[m.to-8]];
	md->victim = game.board[m.to-8];
	game.board[m.to-8] = none;
	game.white_pawns--;
	
      }else{
	game.black_material -= PAWN_VALUE;
	BPawns ^= mask[m.to+8];
	game.b_pieces ^= mask[m.to+8];
	game.zob_key ^= zobrist[PAWN-1][BLACK][m.to+8];
	game.pawn_key ^= zobrist[PAWN-1][BLACK][m.to+8];
	game.all_r_90 ^= mask[r_90_map[m.to+8]];
	game.all_odd_diag ^= mask[odd_diag_map[m.to+8]];
	game.all_even_diag ^= mask[even_diag_map[m.to+8]];
	md->victim = game.board[m.to+8];
	game.board[m.to+8] = none;
	game.black_pawns--;
      }
      /* regular capture code */
    }else{
      switch (game.board[m.to]){
      case w_pawn:
	game.pawn_key ^= zobrist[PAWN-1][WHITE][m.to];
	game.white_pawns--;
	break;
      case b_pawn:
	game.black_pawns--;
	game.pawn_key ^= zobrist[PAWN-1][BLACK][m.to];
	break;
      case w_knight:
	game.white_minors--;
	break;
      case b_knight:
	game.black_minors--;
	break;
      case w_bishop:
	game.white_minors--;
	break;
      case b_bishop:
	game.black_minors--;
	break;
      case  w_rook:
	if(m.to == 56){
	  if(game.new_castle & W_BIG_CASTLE){
#ifdef _DEBUG
	    if(castle_available(W_BIG_CASTLE) == 0)
	      error("castle not really available");
#endif	    
	    game.zob_key ^= castle_zob[WHITE][1];
	    game.new_castle ^= W_BIG_CASTLE;
	  }
	}
	else if(m.to == 63){
	  if(game.new_castle & W_SMALL_CASTLE){
#ifdef _DEBUG
	    if(castle_available(W_SMALL_CASTLE) == 0)
	      error("castle not really available");
#endif	    
	    game.zob_key ^= castle_zob[WHITE][0];
	    game.new_castle ^= W_SMALL_CASTLE;
	  }
	}
	game.white_majors--;
	break;
      case b_rook:
	if(m.to == 0){
	  if(game.new_castle & B_BIG_CASTLE){
#ifdef _DEBUG
	    if(castle_available(B_BIG_CASTLE) == 0)
	      error("castle not really available");
#endif	    
	    game.zob_key ^= castle_zob[BLACK][1];
	    game.new_castle ^= B_BIG_CASTLE;
	  }
	}
	else if(m.to == 7){
	  if(game.new_castle & B_SMALL_CASTLE){
#ifdef _DEBUG
	    if(castle_available(B_SMALL_CASTLE) == 0)
	      error("castle not really available");
#endif	    
	    game.zob_key ^= castle_zob[BLACK][0];
	    game.new_castle ^= B_SMALL_CASTLE;
	  }
	}
	game.black_majors--;
	break;
      case w_queen:
	game.white_majors--;
	break;
      case b_queen:
	game.black_majors--;
	break;
      case w_king:
	game.white_king = -1;
	break;
      case b_king:
	game.black_king = -1;
	break;
      default:
	error("Funny looking piece type");
      }
      game.pieces[game.board[m.to]+6] ^= mask[m.to];
      if(game.board[m.to] < 0){
	game.black_material -= piece_value[game.board[m.to]+6];
	game.b_pieces ^= mask[m.to];
	game.zob_key ^= zobrist[-game.board[m.to]-1][BLACK][m.to];
      }else{
	game.white_material -= piece_value[game.board[m.to]+6];
	game.w_pieces ^= mask[m.to];
	game.zob_key ^= zobrist[game.board[m.to]-1][WHITE][m.to];
      }      
      md->victim = game.board[m.to];
    }
  }else{
    md->victim = none;
  }

  /* end of capture code */
  
  /* move code */
  if(!(m.special & (W_BIG_CASTLE|W_SMALL_CASTLE|B_BIG_CASTLE|B_SMALL_CASTLE))){
      game.pieces[game.board[m.from]+6] ^= mask[m.from];
      game.pieces[game.board[m.from]+6] |= mask[m.to];
    switch(game.board[m.from]){
    case w_pawn:
      game.pawn_key ^= zobrist[PAWN-1][WHITE][m.to];
      game.pawn_key ^= zobrist[PAWN-1][WHITE][m.from];
      break;
    case b_pawn:
      game.pawn_key ^= zobrist[PAWN-1][BLACK][m.to];
      game.pawn_key ^= zobrist[PAWN-1][BLACK][m.from];
      break;
    case w_rook:
      if(m.from == 56){
	if(game.new_castle & W_BIG_CASTLE){
#ifdef _DEBUG
	  if(castle_available(W_BIG_CASTLE) == 0)
	    error("castle not really available");
#endif	    
	  game.zob_key ^= castle_zob[WHITE][1];
	  game.new_castle ^= W_BIG_CASTLE;
	}
      }else if(m.from == 63){
	if(game.new_castle & W_SMALL_CASTLE){
#ifdef _DEBUG
	  if(castle_available(W_SMALL_CASTLE) == 0)
	    error("castle not really available");
#endif	    
	  game.zob_key ^= castle_zob[WHITE][0];
	  game.new_castle ^= W_SMALL_CASTLE;
	}
      }
      break;
    case b_rook:
      if(m.from == 0){
	if(game.new_castle & B_BIG_CASTLE){
#ifdef _DEBUG
	  if(castle_available(B_BIG_CASTLE) == 0)
	    error("castle not really available");
#endif	    
	  game.zob_key ^= castle_zob[BLACK][1];
	  game.new_castle ^= B_BIG_CASTLE;
	}
      }else if(m.from == 7){
	if(game.new_castle & B_SMALL_CASTLE){
#ifdef _DEBUG
	  if(castle_available(B_SMALL_CASTLE) == 0)
	    error("castle not really available");
#endif	    
	  game.zob_key ^= castle_zob[BLACK][0];
	  game.new_castle ^= B_SMALL_CASTLE;
	}
      }
      break;
    case w_king:
      game.white_king = m.to;
      if(game.new_castle & W_SMALL_CASTLE){
	game.zob_key ^= castle_zob[WHITE][0];
	game.new_castle ^= W_SMALL_CASTLE;
      }
      if(game.new_castle & W_BIG_CASTLE){
	game.zob_key ^= castle_zob[WHITE][1];      
	game.new_castle ^= W_BIG_CASTLE;
      }

      break;
    case b_king:
      game.black_king = m.to;
      if(game.new_castle & B_SMALL_CASTLE){
	game.zob_key ^= castle_zob[BLACK][0];
	game.new_castle ^= B_SMALL_CASTLE;      
      }
      if(game.new_castle & B_BIG_CASTLE){
	game.zob_key ^= castle_zob[BLACK][1];   
	game.new_castle ^= B_BIG_CASTLE;         
      }
      break;
    }
    
    if(game.board[m.from] < 0){
      game.b_pieces ^= mask[m.from];
      game.b_pieces |= mask[m.to];
      game.zob_key ^= zobrist[-game.board[m.from]-1][BLACK][m.from];
      game.zob_key ^= zobrist[-game.board[m.from]-1][BLACK][m.to];
    }else{
      game.w_pieces ^= mask[m.from];
      game.w_pieces |= mask[m.to];
      game.zob_key ^= zobrist[game.board[m.from]-1][WHITE][m.from];
      game.zob_key ^= zobrist[game.board[m.from]-1][WHITE][m.to];
    }      

    game.all_r_90 ^= mask[r_90_map[m.from]];
    game.all_r_90 |= mask[r_90_map[m.to]];
    game.all_odd_diag ^= mask[odd_diag_map[m.from]];
    game.all_odd_diag |= mask[odd_diag_map[m.to]];
    game.all_even_diag ^= mask[even_diag_map[m.from]];
    game.all_even_diag |= mask[even_diag_map[m.to]];
    
    game.board[m.to] = game.board[m.from];
    game.board[m.from] = none;
  }else{
    /* and now the castling code */
    if(m.special & W_BIG_CASTLE){
      sq1 = 56;
      sq2 = 60;
      sq3 = 59;
      sq4 = 58;
    }else if(m.special & W_SMALL_CASTLE){
      sq1 = 63;
      sq2 = 60;
      sq3 = 61;
      sq4 = 62;
    }else if(m.special & B_BIG_CASTLE){
      sq1 = 0;
      sq2 = 4;
      sq3 = 3;
      sq4 = 2;
    }else if(m.special & B_SMALL_CASTLE){
      sq1 = 7;
      sq2 = 4;
      sq3 = 5;
      sq4 = 6;
    }

    if(m.special & (W_BIG_CASTLE | W_SMALL_CASTLE)){
      game.board[sq2] = none;
      game.board[sq1] = none;
      game.board[sq4] = w_king;
      game.board[sq3] = w_rook;

      WRooks ^= mask[sq1];
      game.w_pieces ^= mask[sq1];
      WRooks |= mask[sq3];
      game.w_pieces |= mask[sq3];
      
      game.white_king = sq4;
      game.w_pieces ^= mask[sq2];
      game.w_pieces |= mask[sq4];

      game.zob_key ^= zobrist[KING-1][WHITE][sq2];
      game.zob_key ^= zobrist[KING-1][WHITE][sq4];
      game.zob_key ^= zobrist[ROOK-1][WHITE][sq1];
      game.zob_key ^= zobrist[ROOK-1][WHITE][sq3];

      if(game.new_castle & W_SMALL_CASTLE)
	game.zob_key ^= castle_zob[WHITE][0];
      if(game.new_castle & W_BIG_CASTLE)
	game.zob_key ^= castle_zob[WHITE][1];      

      game.new_castle &= ~(W_BIG_CASTLE|W_SMALL_CASTLE);
      game.new_castle |= W_CASTLED;

    }else{
      game.board[sq2] = none;
      game.board[sq1] = none;
      game.board[sq4] = b_king;
      game.board[sq3] = b_rook;      
      
      BRooks ^= mask[sq1];
      game.b_pieces ^= mask[sq1];
      BRooks |= mask[sq3];
      game.b_pieces |= mask[sq3];
      
      game.black_king = sq4;
      game.b_pieces ^= mask[sq2];
      game.b_pieces |= mask[sq4];
      game.zob_key ^= zobrist[KING-1][BLACK][sq2];
      game.zob_key ^= zobrist[KING-1][BLACK][sq4];
      game.zob_key ^= zobrist[ROOK-1][BLACK][sq1];
      game.zob_key ^= zobrist[ROOK-1][BLACK][sq3];
      if(game.new_castle & B_SMALL_CASTLE)
	game.zob_key ^= castle_zob[BLACK][0];
      if(game.new_castle & B_BIG_CASTLE)
	game.zob_key ^= castle_zob[BLACK][1];      

      game.new_castle &= ~(B_BIG_CASTLE|B_SMALL_CASTLE);
      game.new_castle |= B_CASTLED;

    }
    game.all_r_90 ^= mask[r_90_map[sq2]];
    game.all_r_90 ^= mask[r_90_map[sq1]];
    game.all_r_90 |= mask[r_90_map[sq4]];
    game.all_r_90 |= mask[r_90_map[sq3]];
    game.all_odd_diag ^= mask[odd_diag_map[sq2]];
    game.all_odd_diag ^= mask[odd_diag_map[sq1]];
    game.all_odd_diag |= mask[odd_diag_map[sq4]];
    game.all_odd_diag |= mask[odd_diag_map[sq3]];
    game.all_even_diag ^= mask[even_diag_map[sq2]];
    game.all_even_diag ^= mask[even_diag_map[sq1]];
    game.all_even_diag |= mask[even_diag_map[sq4]];
    game.all_even_diag |= mask[even_diag_map[sq3]];
        
  }
  /* Irreversability code */
  if(m.special & IRREVERSIBLE){

    history_floor += move_50_counter;
    
    move_50_counter = 0;

    /* Promotion and en passant are irreversible moves*/

    /* Promotion code */
    if(m.to_promote){
      if(game.board[m.to] > 0){
	game.board[m.to] = m.to_promote;
      }else{
	game.board[m.to] = -m.to_promote;
      }
      game.pieces[game.board[m.to]+6] |= mask[m.to];
      
      if(game.board[m.to] < 0){
	switch(game.board[m.to]){
	case b_queen:
	case b_rook:
	  game.black_majors++;
	  break;
	case b_knight:
	case b_bishop:
	  game.black_minors++;
	  break;
	}
	game.black_pawns--;
	BPawns ^= mask[m.to];
	game.pawn_key ^= zobrist[PAWN-1][BLACK][m.to];
	game.black_material += piece_value[game.board[m.to]+6]-PAWN_VALUE;
	game.zob_key ^= zobrist[PAWN-1][BLACK][m.to];
	game.zob_key ^= zobrist[-game.board[m.to]-1][BLACK][m.to];
      }else{
	switch(game.board[m.to]){
	case w_queen:
	case w_rook:
	  game.white_majors++;
	  break;
	case w_knight:
	case w_bishop:
	  game.white_minors++;
	  break;
	}
	game.white_pawns--;
	WPawns ^= mask[m.to];
	game.pawn_key ^= zobrist[PAWN-1][WHITE][m.to];
	game.white_material += piece_value[game.board[m.to]+6]-PAWN_VALUE;
	game.zob_key ^= zobrist[PAWN-1][WHITE][m.to];
	game.zob_key ^= zobrist[game.board[m.to]-1][WHITE][m.to];
      }
    }
    /* possible en passant code */
    if(m.special & POSSIBLE_EN_PASSANT){
      /* only update the en_passant target if there is a possible capture */
      if(game.board[m.to] > 0 && (en_passant_mask[m.to+8] & BPawns)){
	game.en_passant = m.to+8;
	game.zob_key ^= en_passant_zob[(int)game.en_passant];
      }else if(en_passant_mask[m.to-8] & WPawns){
	game.en_passant = m.to-8;
	game.zob_key ^= en_passant_zob[(int)game.en_passant];
      }
    }
  }
  
  /* Change player */
  /* This part of the key is not in crafty because i suppose it
   uses 2 separate hash tables, one for each color. So when using the
  hash for the book i have to XOR change_side_zob if the side to move is black */
  game.zob_key ^= change_side_zob;
  game.to_move = !game.to_move;

  /* move data stack pointer */
  data_stack_top++;


  /* somehow i think placing the zob on the history stack should
   only be made AFTER all changes to the zob key are done!!*/
  /* History Code */
  *(history_floor+move_50_counter) = game.zob_key;



#ifdef _DEBUG
  
  if(!check_boards_cons()){
    print_board(stdout);
    print_moves(&m+1,&m);
    error("Insane move!");
  }
  if(gen_zob_key() != game.zob_key || gen_pawn_key() != game.pawn_key){
    print_board(stdout);
    print_moves(&m+1,&m);
    error("Zobrist will be really mad!");
  }
  
#endif
}

void undo_move(Move m){      
  Move_Data * md;
  int sq1 = 0;
  int sq2 = 0;
  int sq3 = 0;
  int sq4 = 0;

#ifdef _DEBUG
  

  if(!check_boards_cons()){
    print_board(stdout);
    print_moves(&m+1,&m);
    error("Insane move!");
  }
  if(gen_zob_key() != game.zob_key || gen_pawn_key() != game.pawn_key){
    print_board(stdout);
    print_moves(&m+1,&m);
    error("Zobrist will be really mad!");
  }
  
#endif

  /* adjust stack_top*/
  data_stack_top--;
  md = data_stack_top;

  /* Change player */
  game.to_move = !game.to_move;

  /* restore history */
  history_floor = md->history_floor;
  move_50_counter = md->move_50_counter;

  /* restore zob key and en_passant */
  game.zob_key = md->zob_key;
  game.pawn_key = md->pawn_key;
  game.en_passant = md->en_passant;
  game.new_castle = md->new_castle;

  /* Promotion code */
  if(m.to_promote){
    /* this takes advantage of the fact that the pawn code
     is 1 or -1 */
    game.pieces[game.board[m.to]+6] ^= mask[m.to];
      
    if(game.board[m.to] < 0){
      switch(game.board[m.to]){
      case b_queen:
      case b_rook:
	game.black_majors--;
	break;
      case b_knight:
      case b_bishop:
	game.black_minors--;
	break;
      }
      game.black_pawns++;
      BPawns ^= mask[m.to];
      game.black_material -= piece_value[game.board[m.to]+6]-PAWN_VALUE;
    }else{
	switch(game.board[m.to]){
	case w_queen:
	case w_rook:
	  game.white_majors++;
	  break;
	case w_knight:
	case w_bishop:
	  game.white_minors++;
	  break;
	}
      game.white_pawns++;
      WPawns ^= mask[m.to];
      game.white_material -= piece_value[game.board[m.to]+6]-PAWN_VALUE;
    }
    if(game.board[m.to] > 0){
      game.board[m.to] = w_pawn;
    }else{
      game.board[m.to] = b_pawn;
    }
  }

  /* move code */
  if(!(m.special & (W_BIG_CASTLE|W_SMALL_CASTLE|B_BIG_CASTLE|B_SMALL_CASTLE))){
    game.pieces[game.board[m.to]+6] ^= mask[m.from];
    game.pieces[game.board[m.to]+6] ^= mask[m.to];
    switch(game.board[m.to]){
    case w_king:
      game.white_king = m.from;
      break;
    case b_king:
      game.black_king = m.from;
      break;
    }
    if(game.board[m.to] < 0){
      game.b_pieces ^= mask[m.from];
      game.b_pieces ^= mask[m.to];
    }else{
      game.w_pieces ^= mask[m.from];
      game.w_pieces ^= mask[m.to];
    }      

    game.all_r_90 ^= mask[r_90_map[m.from]];
    game.all_r_90 ^= mask[r_90_map[m.to]];
    game.all_odd_diag ^= mask[odd_diag_map[m.from]];
    game.all_odd_diag ^= mask[odd_diag_map[m.to]];
    game.all_even_diag ^= mask[even_diag_map[m.from]];
    game.all_even_diag ^= mask[even_diag_map[m.to]];
    
    game.board[m.from] = game.board[m.to];
    game.board[m.to] = none;
  }else{
    /* and now the castling code */
    if(m.special & W_BIG_CASTLE){
      sq1 = 56;
      sq2 = 60;
      sq3 = 59;
      sq4 = 58;
    }else if(m.special & W_SMALL_CASTLE){
      sq1 = 63;
      sq2 = 60;
      sq3 = 61;
      sq4 = 62;
    }else if(m.special & B_BIG_CASTLE){
      sq1 = 0;
      sq2 = 4;
      sq3 = 3;
      sq4 = 2;
    }else if(m.special & B_SMALL_CASTLE){
      sq1 = 7;
      sq2 = 4;
      sq3 = 5;
      sq4 = 6;
    }

    if(m.special & (W_BIG_CASTLE|W_SMALL_CASTLE)){
      game.board[sq2] = w_king;
      game.board[sq1] = w_rook;
      game.board[sq4] = none;
      game.board[sq3] = none;

      WRooks ^= mask[sq1];
      game.w_pieces ^= mask[sq1];
      WRooks ^= mask[sq3];
      game.w_pieces ^= mask[sq3];
      
      game.white_king = sq2;
      game.w_pieces ^= mask[sq2];
      game.w_pieces ^= mask[sq4];
    }else{
      game.board[sq2] = b_king;
      game.board[sq1] = b_rook;
      game.board[sq4] = none;
      game.board[sq3] = none;

      BRooks ^= mask[sq1];
      game.b_pieces ^= mask[sq1];
      BRooks ^= mask[sq3];
      game.b_pieces ^= mask[sq3];
      
      game.black_king = sq2;
      game.b_pieces ^= mask[sq2];
      game.b_pieces ^= mask[sq4];
    }

    game.all_r_90 ^= mask[r_90_map[sq2]];
    game.all_r_90 ^= mask[r_90_map[sq1]];
    game.all_r_90 ^= mask[r_90_map[sq4]];
    game.all_r_90 ^= mask[r_90_map[sq3]];
    game.all_odd_diag ^= mask[odd_diag_map[sq2]];
    game.all_odd_diag ^= mask[odd_diag_map[sq1]];
    game.all_odd_diag ^= mask[odd_diag_map[sq4]];
    game.all_odd_diag ^= mask[odd_diag_map[sq3]];
    game.all_even_diag ^= mask[even_diag_map[sq2]];
    game.all_even_diag ^= mask[even_diag_map[sq1]];
    game.all_even_diag ^= mask[even_diag_map[sq4]];
    game.all_even_diag ^= mask[even_diag_map[sq3]];
        
  }

  /* capture code */
  if(md->victim){
    /* en passant code*/
    if(m.special & EN_PASSANT){
      if(game.board[m.from] < 0){
	/*target is white */
	game.white_material += PAWN_VALUE;
	game.white_pawns++;
	WPawns ^= mask[m.to-8];
	game.w_pieces ^= mask[m.to-8];
	game.all_r_90 ^= mask[r_90_map[m.to-8]];
	game.all_odd_diag ^= mask[odd_diag_map[m.to-8]];
	game.all_even_diag ^= mask[even_diag_map[m.to-8]];
	game.board[m.to-8] = w_pawn;
      }else{
	game.black_material += PAWN_VALUE;
	game.black_pawns++;
	BPawns ^= mask[m.to+8];
	game.b_pieces ^= mask[m.to+8];
	game.all_r_90 ^= mask[r_90_map[m.to+8]];
	game.all_odd_diag ^= mask[odd_diag_map[m.to+8]];
	game.all_even_diag ^= mask[even_diag_map[m.to+8]];
	game.board[m.to+8] = b_pawn;
      }
      /* regular capture code */
    }else{
      switch(md->victim){
      case w_bishop:
      case w_knight:
	game.white_minors++;
	break;
      case w_rook:
      case w_queen:
	game.white_majors++;
	break;
      case b_bishop:
      case b_knight:
	game.black_minors++;
	break;
      case b_rook:
      case b_queen:
	game.black_majors++;
	break;
      case w_king:
	game.white_king = m.to;
	break;
      case b_king:
	game.black_king = m.to;
	break;
      case w_pawn:
	game.white_pawns++;
	break;
      case b_pawn:
	game.black_pawns++;
	break;
      }
      game.pieces[md->victim+6] ^= mask[m.to];
      if(md->victim < 0){
	game.black_material += piece_value[md->victim+6];
	game.b_pieces ^= mask[m.to];
      }else if(md->victim > 0){
	game.white_material += piece_value[md->victim+6];
	game.w_pieces ^= mask[m.to];
      }else{
	error("Unkown victim");
      }
      game.all_r_90 ^= mask[r_90_map[m.to]];
      game.all_odd_diag ^= mask[odd_diag_map[m.to]];
      game.all_even_diag ^= mask[even_diag_map[m.to]];
      game.board[m.to] = md->victim;
    }
  }

#ifdef _DEBUG
  if(!check_boards_cons()){
    print_board(stdout);
    print_moves(&m+1,&m);
    error("Insane move!");
  }
  if(gen_zob_key() != game.zob_key || gen_pawn_key() != game.pawn_key){
    print_board(stdout);
    print_moves(&m+1,&m);
    error("Zobrist will be really mad!");
  }
#endif
}

int castle_available(int flag){
  return flag & game.new_castle;
}
